#
# Copyright (c) 2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        acl.py
#
# Abstract:
#
#        SMB2/3 ACL set/get testing.
#
#

import pike.model
import pike.smb2
import pike.test
import pike.ntstatus

from pike.test import unittest


class SecTest(pike.test.PikeTest):
    def __init__(self, *args, **kwargs):
        super(SecTest, self).__init__(*args, **kwargs)
        self.chan = None
        self.tree = None
        self.sec_info = (
            pike.smb2.OWNER_SECURITY_INFORMATION
            | pike.smb2.GROUP_SECURITY_INFORMATION
            | pike.smb2.DACL_SECURITY_INFORMATION
        )
        self.sec_info_owner = pike.smb2.OWNER_SECURITY_INFORMATION
        self.sec_info_group = pike.smb2.GROUP_SECURITY_INFORMATION
        self.sec_info_dacl = pike.smb2.DACL_SECURITY_INFORMATION

    def open_file(self):
        """Helper to open basic file"""
        self.chan, self.tree = self.tree_connect()
        return self.chan.create(
            self.tree,
            "test_sec.txt",
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            access=(
                pike.smb2.READ_CONTROL
                | pike.smb2.WRITE_OWNER
                | pike.smb2.DELETE
                | pike.smb2.WRITE_DAC
            ),
            disposition=pike.smb2.FILE_SUPERSEDE,
        ).result()

    def test_set_sec_same(self):
        handle = self.open_file()
        info = self.chan.query_file_info(
            handle,
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=self.sec_info,
        )
        with self.chan.set_file_info(
            handle,
            pike.smb2.FileSecurityInformation,
            additional_information=self.sec_info,
        ) as file_info:
            file_info.clone_from(info)
        self.chan.close(handle)

    def test_set_sec_owner(self):
        handle = self.open_file()
        info = self.chan.query_file_info(
            handle,
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=self.sec_info,
        )
        with self.chan.set_file_info(
            handle,
            pike.smb2.FileSecurityInformation,
            additional_information=self.sec_info_owner,
        ) as file_info:
            file_info.clone_from(info, ["owner_sid"])
        self.chan.close(handle)

    def test_set_sec_group(self):
        handle = self.open_file()
        info = self.chan.query_file_info(
            handle,
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=self.sec_info,
        )
        with self.chan.set_file_info(
            handle,
            pike.smb2.FileSecurityInformation,
            additional_information=self.sec_info_group,
        ) as file_info:
            file_info.clone_from(info, ["group_sid"])
        self.chan.close(handle)

    def test_set_sec_dacl(self):
        handle = self.open_file()
        info = self.chan.query_file_info(
            handle,
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=self.sec_info,
        )
        with self.chan.set_file_info(
            handle,
            pike.smb2.FileSecurityInformation,
            additional_information=self.sec_info_dacl,
        ) as file_info:
            file_info.clone_from(info, ["dacl"])
        self.chan.close(handle)

    def test_set_sec_dacl_partial_copy(self):
        handle = self.open_file()
        info = self.chan.query_file_info(
            handle,
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=self.sec_info,
        )
        with self.chan.set_file_info(
            handle,
            pike.smb2.FileSecurityInformation,
            additional_information=self.sec_info_dacl,
        ) as file_info:
            file_info.clone_from(info, [])
            aces = info.dacl.children
            new_dacl = pike.smb2.NT_ACL()
            new_dacl.acl_revision = pike.smb2.ACL_REVISION
            new_dacl.append(aces[0])
            file_info.dacl = new_dacl
        self.chan.close(handle)

    def test_set_sec_dacl_append_ace(self):
        handle = self.open_file()
        info = self.chan.query_file_info(
            handle,
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=self.sec_info,
        )
        with self.chan.set_file_info(
            handle,
            pike.smb2.FileSecurityInformation,
            additional_information=self.sec_info_dacl,
        ) as file_info:
            file_info.clone_from(info, [])
            aces = info.dacl.children
            new_dacl = info.dacl.clone()
            file_info.dacl = new_dacl
            # append a modified ace
            template_ace = aces[0].clone(new_parent=None)
            template_ace.sid.identifier_authority = 5
            template_ace.sid.sub_authority = [18]
            file_info.dacl.append(template_ace)
        self.chan.close(handle)

    def test_set_sec_dacl_new(self):
        handle = self.open_file()
        info = self.chan.query_file_info(
            handle,
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=self.sec_info,
        )
        with self.chan.set_file_info(
            handle,
            pike.smb2.FileSecurityInformation,
            additional_information=self.sec_info_dacl,
        ) as file_info:
            file_info.clone_from(info, [])
            aces = info.dacl.children
            # generate a new NT_ACL object
            new_dacl = pike.smb2.NT_ACL()
            new_dacl.acl_revision = pike.smb2.ACL_REVISION
            # keep the original aces
            new_dacl._entries = aces
            # append a cloned ace then modify it
            template_ace = aces[0].clone(new_parent=new_dacl)
            template_ace.sid.identifier_authority = 5
            template_ace.sid.sub_authority = [18]
            file_info.dacl = new_dacl
        self.chan.close(handle)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        appinstanceid.py
#
# Abstract:
#
#        App instance ID tests
#
# Authors: Arlene Berry (arlene.berry@emc.com)
#

from builtins import map
import pike.model
import pike.smb2
import pike.test
import pike.ntstatus
import random
import array


@pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
@pike.test.RequireCapabilities(pike.smb2.SMB2_GLOBAL_CAP_PERSISTENT_HANDLES)
@pike.test.RequireShareCapabilities(pike.smb2.SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY)
class AppInstanceIdTest(pike.test.PikeTest):
    def __init__(self, *args, **kwargs):
        super(AppInstanceIdTest, self).__init__(*args, **kwargs)
        self.share_all = (
            pike.smb2.FILE_SHARE_READ
            | pike.smb2.FILE_SHARE_WRITE
            | pike.smb2.FILE_SHARE_DELETE
        )
        self.lease1 = array.array("B", map(random.randint, [0] * 16, [255] * 16))
        self.app_instance_id1 = array.array(
            "B", map(random.randint, [0] * 16, [255] * 16)
        )
        self.r = pike.smb2.SMB2_LEASE_READ_CACHING
        self.rw = self.r | pike.smb2.SMB2_LEASE_WRITE_CACHING
        self.rh = self.r | pike.smb2.SMB2_LEASE_HANDLE_CACHING
        self.rwh = self.rw | self.rh

    # Perform create with AppInstanceId
    def test_appinstanceid(self):
        chan, tree = self.tree_connect()

        # Request rwh lease
        handle1 = chan.create(
            tree,
            "appinstanceid.txt",
            access=pike.smb2.FILE_READ_DATA
            | pike.smb2.FILE_WRITE_DATA
            | pike.smb2.DELETE,
            share=self.share_all,
            disposition=pike.smb2.FILE_OPEN_IF,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
            durable=0,
            persistent=True,
            app_instance_id=self.app_instance_id1,
        ).result()

        self.assertEqual(handle1.durable_flags, pike.smb2.SMB2_DHANDLE_FLAG_PERSISTENT)
        self.assertEqual(handle1.lease.lease_state, self.rwh)

        chan.close(handle1)

    # Invalidate disconnected persistent handle via AppInstanceId
    def test_appinstanceid_persistent_with_disconnect(self):
        chan, tree = self.tree_connect()

        # Request rwh lease
        handle1 = chan.create(
            tree,
            "appinstanceid.txt",
            share=self.share_all,
            disposition=pike.smb2.FILE_OPEN_IF,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
            durable=0,
            persistent=True,
            app_instance_id=self.app_instance_id1,
        ).result()

        self.assertEqual(handle1.durable_flags, pike.smb2.SMB2_DHANDLE_FLAG_PERSISTENT)
        self.assertEqual(handle1.lease.lease_state, self.rwh)

        # Close the connection
        chan.connection.close()

        chan2, tree2 = self.tree_connect(client=pike.model.Client())

        # Request reconnect
        handle2 = chan2.create(
            tree,
            "appinstanceid.txt",
            access=pike.smb2.FILE_READ_DATA
            | pike.smb2.FILE_WRITE_DATA
            | pike.smb2.DELETE,
            share=self.share_all,
            disposition=pike.smb2.FILE_OPEN_IF,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
            durable=0,
            persistent=True,
            app_instance_id=self.app_instance_id1,
        ).result()

        self.assertEqual(handle2.durable_flags, pike.smb2.SMB2_DHANDLE_FLAG_PERSISTENT)
        self.assertEqual(handle2.lease.lease_state, self.rwh)

        chan2.close(handle2)

    # Invalidating a disconnected persistent handle
    # with AppInstanceId fails when reusing the same client GUID
    def test_appinstanceid_reconnect_same_clientguid(self):
        chan, tree = self.tree_connect()

        # Request rwh lease
        handle1 = chan.create(
            tree,
            "appinstanceid.txt",
            share=self.share_all,
            disposition=pike.smb2.FILE_OPEN_IF,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
            durable=0,
            persistent=True,
            app_instance_id=self.app_instance_id1,
        ).result()

        self.assertEqual(handle1.durable_flags, pike.smb2.SMB2_DHANDLE_FLAG_PERSISTENT)
        self.assertEqual(handle1.lease.lease_state, self.rwh)

        # Close the connection
        chan.connection.close()

        chan2, tree2 = self.tree_connect()

        # Request reconnect
        with self.assert_error(pike.ntstatus.STATUS_FILE_NOT_AVAILABLE):
            handle2 = chan2.create(
                tree,
                "appinstanceid.txt",
                access=pike.smb2.FILE_READ_DATA
                | pike.smb2.FILE_WRITE_DATA
                | pike.smb2.DELETE,
                share=self.share_all,
                disposition=pike.smb2.FILE_OPEN_IF,
                options=pike.smb2.FILE_DELETE_ON_CLOSE,
                oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
                lease_key=self.lease1,
                lease_state=self.rwh,
                durable=0,
                persistent=True,
                app_instance_id=self.app_instance_id1,
            ).result()

        # Close the connection
        chan2.connection.close()

        chan3, tree3 = self.tree_connect(client=pike.model.Client())

        # Request reconnect
        handle3 = chan3.create(
            tree,
            "appinstanceid.txt",
            access=pike.smb2.FILE_READ_DATA
            | pike.smb2.FILE_WRITE_DATA
            | pike.smb2.DELETE,
            share=self.share_all,
            disposition=pike.smb2.FILE_OPEN_IF,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease1,
            lease_state=self.rwh,
            durable=0,
            persistent=True,
            app_instance_id=self.app_instance_id1,
        ).result()

        self.assertEqual(handle3.durable_flags, pike.smb2.SMB2_DHANDLE_FLAG_PERSISTENT)
        self.assertEqual(handle3.lease.lease_state, self.rwh)

        chan3.close(handle3)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        nttime.py
#
# Abstract:
#
#        NT time format utility class
#
# Authors: Rafal Szczesniak (rafal.szczesniak@isilon.com)
#          Masen Furer (masen.furer@dell.com)
#

from __future__ import division
from builtins import str
from past.builtins import basestring

from datetime import datetime, timedelta
import math
import time

_unix_time_offset = 11644473600
_unix_epoch = datetime.fromtimestamp(0) + timedelta(hours=time.localtime().tm_isdst)
_intervals_per_second = 10000000


def _unix_time_to_nt_time(t):
    return (t + _unix_time_offset) * _intervals_per_second


def _datetime_to_unix_time(t):
    td = t - _unix_epoch
    return td.seconds + (td.days * 24 * 3600)


def _datetime_to_nt_time(t):
    return _unix_time_to_nt_time(_datetime_to_unix_time(t))


def _nt_time_to_unix_time(t):
    """
    Converts "FILETIME" to Python's time library format.

    routine borrowed from dist_shell: sys_ex/windows/file.py
    """

    # This calculation below might seem rather complex for such time
    # conversions, however, due to issues with floating point
    # arithmetics precision, this HACK is required to compute the time
    # in Python's time library format such that times like "12:59.59.1"
    # don't turn into "01:00:00.0" when converted back. This
    # calculation below makes use of Python's bigint capabilities.
    py_time = (t << 32) // _intervals_per_second
    py_time -= _unix_time_offset << 32
    py_time_parts = divmod(py_time, 2**32)
    py_time = float(py_time_parts[0])
    py_time += math.copysign(py_time_parts[1], py_time) // (2**32)
    return py_time


def GMT_to_datetime(gmt_token):
    dt_obj = datetime.strptime(gmt_token, "@GMT-%Y.%m.%d-%H.%M.%S")
    # apply timezone conversion
    dt_obj -= timedelta(seconds=time.timezone)
    return dt_obj


class NtTime(int):
    """
    NtTime may be initialized with any of the following values
      * string in ISO format or @GMT- format (timewarp)
      * datetime
      * int/long indicating number of 100-nanosecond intervals since 1601-01-01
    """

    def __new__(cls, value):
        if isinstance(value, basestring):
            if value.startswith("@GMT-"):
                dt = GMT_to_datetime(value)
            else:
                dt = datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
            value = _datetime_to_nt_time(dt)
        elif isinstance(value, datetime):
            value = _datetime_to_nt_time(value)

        return super(NtTime, cls).__new__(cls, int(value))

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.to_datetime().isoformat(" ")

    def to_datetime(self):
        return datetime.fromtimestamp(_nt_time_to_unix_time(self))

    def to_pytime(self):
        return _nt_time_to_unix_time(self)

    def to_unixtime(self):
        return int(self.to_pytime())



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        ntlm.py
#
# Abstract:
#
#        Pure python NTLM authentication provider
#
# Authors: Masen Furer (masen.furer@emc.com)
#

from __future__ import absolute_import
from builtins import object
from builtins import range

import array
import random
from socket import gethostname
import struct

try:
    import Crypto.Cipher.DES
    import Crypto.Cipher.ARC4 as RC4
    import Crypto.Hash.HMAC as HMAC
    import Crypto.Hash.MD4 as MD4
    import Crypto.Hash.MD5 as MD5
except ImportError:
    import Cryptodome.Cipher.DES
    import Cryptodome.Cipher.ARC4 as RC4
    import Cryptodome.Hash.HMAC as HMAC
    import Cryptodome.Hash.MD4 as MD4
    import Cryptodome.Hash.MD5 as MD5

from . import core
from . import nttime


def des_key_64(K):
    """
    Return an 64-bit des key by adding a zero to the least significant bit
    of each character.
    K should be a 7 char string
    """
    in_key = K + "\0"
    out_key = array.array("B", K[0])
    for ix in range(1, len(in_key)):
        out_key.append(
            ((ord(in_key[ix - 1]) << (8 - ix)) & 0xFF) | (ord(in_key[ix]) >> ix)
        )
    return out_key.tobytes()


def DES(K, D):
    d1 = Cryptodome.Cipher.DES.new(des_key_64(K), Cryptodome.Cipher.DES.MODE_ECB)
    return d1.encrypt(D)


def DESL(K, D):
    return DES(K[:7], D) + DES(K[7:14], D) + DES(K[14:16] + "\0" * 5, D)


def nonce(length):
    return array.array("B", [random.getrandbits(8) for x in range(length)])


def encode_frame(frame):
    buffer = array.array("B")
    frame.encode(core.Cursor(buffer, 0))
    return buffer


NTLM_SIGNATURE = b"NTLMSSP\x00"


class MessageType(core.ValueEnum):
    NtLmNegotiate = 0x1
    NtLmChallenge = 0x2
    NtLmAuthenticate = 0x3


MessageType.import_items(globals())


class Ntlm(core.Frame):
    def __init__(self, parent=None):
        core.Frame.__init__(self, parent)
        self.message = None

    def _encode(self, cur):
        cur.encode_bytes(NTLM_SIGNATURE)
        if self.message is not None:
            self.message.encode(cur)

    def _decode(self, cur):
        signature = cur.decode_bytes(8)
        if signature.tobytes() != NTLM_SIGNATURE:
            raise core.BadPacket("Packet signature does not match")
        # determine the message type to pass off decoding
        message_type = cur.decode_uint32le()
        if message_type == NtLmChallenge:
            self.message = NtLmChallengeMessage(self)
            self.message.decode(cur)
        else:
            raise core.BadPacket(
                "Unknown response message type: {0}".format(message_type)
            )


class NegotiateFlags(core.FlagEnum):
    NTLMSSP_NEGOTIATE_UNICODE = 0x00000001
    NTLM_NEGOTIATE_OEM = 0x00000002
    NTLMSSP_REQUEST_TARGET = 0x00000004
    NTLM_NEGOTIATE_SIGN = 0x00000010
    NTLM_NEGOTIATE_SEAL = 0x00000020
    NTLMSSP_NEGOTIATE_DATAGRAM = 0x00000040
    NTLMSSP_NEGOTIATE_LM_KEY = 0x00000080
    NTLMSSP_NEGOTIATE_NTLM = 0x00000200
    ANONYMOUS = 0x00000800
    NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED = 0x00001000
    NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED = 0x00002000
    NTLMSSP_NEGOTIATE_ALWAYS_SIGN = 0x00008000
    NTLMSSP_TARGET_TYPE_DOMAIN = 0x00010000
    NTLMSSP_TARGET_TYPE_SERVER = 0x00020000
    NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY = 0x00080000
    NTLMSSP_NEGOTIATE_IDENTIFY = 0x00100000
    NTLMSSP_REQUEST_NON_NT_SESSION_KEY = 0x00400000
    NTLMSSP_NEGOTIATE_TARGET_INFO = 0x00800000
    NTLMSSP_NEGOTIATE_VERSION = 0x02000000
    NTLMSSP_NEGOTIATE_128 = 0x20000000
    NTLMSSP_NEGOTIATE_KEY_EXCH = 0x40000000
    NTLMSSP_NEGOTIATE_56 = 0x80000000


NegotiateFlags.import_items(globals())


class ProductMajorVersionFlags(core.ValueEnum):
    WINDOWS_MAJOR_VERSION_5 = 0x5
    WINDOWS_MAJOR_VERSION_6 = 0x6
    WINDOWS_MAJOR_VERSION_10 = 0xA


ProductMajorVersionFlags.import_items(globals())


class ProductMinorVersionFlags(core.ValueEnum):
    WINDOWS_MINOR_VERSION_0 = 0x0
    WINDOWS_MINOR_VERSION_1 = 0x1
    WINDOWS_MINOR_VERSION_2 = 0x2
    WINDOWS_MINOR_VERSION_3 = 0x3


ProductMinorVersionFlags.import_items(globals())


class NTLMRevisionCurrentFlags(core.ValueEnum):
    UNKNOWN = 0x0
    NTLMSSP_REVISION_W2K3 = 0x0F


NTLMRevisionCurrentFlags.import_items(globals())


class Version(core.Frame):
    def __init__(self, parent=None):
        core.Frame.__init__(self, parent)
        if parent is not None:
            parent.version = self
        self.product_major_version = WINDOWS_MAJOR_VERSION_5
        self.product_minor_version = WINDOWS_MINOR_VERSION_0
        self.product_build = 0
        self.ntlm_revision_current = NTLMSSP_REVISION_W2K3

    def _encode(self, cur):
        cur.encode_uint8le(self.product_major_version)
        cur.encode_uint8le(self.product_minor_version)
        cur.encode_uint16le(self.product_build)
        cur.encode_uint16le(0)  # reserved
        cur.encode_uint8le(0)  # reserved
        cur.encode_uint8le(self.ntlm_revision_current)

    def _decode(self, cur):
        self.product_major_version = ProductMajorVersionFlags(cur.decode_uint8le())
        self.product_minor_version = ProductMinorVersionFlags(cur.decode_uint8le())
        self.product_build = cur.decode_uint16le()
        reserved = cur.decode_uint16le()
        reserved = cur.decode_uint8le()
        self.ntlm_revision_current = NTLMRevisionCurrentFlags(cur.decode_uint8le())


class NtLmNegotiateMessage(core.Frame):
    message_type = NtLmNegotiate

    def __init__(self, parent=None):
        core.Frame.__init__(self, parent)
        if parent is not None:
            parent.message = self
        self.negotiate_flags = 0
        self.domain_name = ""
        self.workstation_name = gethostname()
        self.version = None

    def _encode(self, cur):
        is_unicode = self.negotiate_flags & NTLMSSP_NEGOTIATE_UNICODE

        message_start = cur - 8
        cur.encode_uint32le(self.message_type)
        cur.encode_uint32le(self.negotiate_flags)

        if self.negotiate_flags & NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED:
            domain_name_len = len(self.domain_name)
        else:
            domain_name_len = 0
        if is_unicode:
            domain_name_len = domain_name_len * 2
        cur.encode_uint16le(domain_name_len)
        cur.encode_uint16le(domain_name_len)
        domain_name_offset_hole = cur.hole.encode_uint32le(0)

        if self.negotiate_flags & NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED:
            workstation_name_len = len(self.workstation_name)
        else:
            workstation_name_len = 0
        if is_unicode:
            workstation_name_len = workstation_name_len * 2
        cur.encode_uint16le(workstation_name_len)
        cur.encode_uint16le(workstation_name_len)
        workstation_name_offset_hole = cur.hole.encode_uint32le(0)

        if (
            self.negotiate_flags & NTLMSSP_NEGOTIATE_VERSION
            and self.version is not None
        ):
            self.version.encode(cur)
        else:
            cur.encode_uint64le(0)

        domain_name_offset_hole(cur - message_start)
        if domain_name_len > 0:
            if is_unicode:
                cur.encode_utf16le(self.domain_name)
            else:
                cur.encode_bytes(self.domain_name)

        workstation_name_offset_hole(cur - message_start)
        if workstation_name_len > 0:
            if is_unicode:
                cur.encode_utf16le(self.workstation_name)
            else:
                cur.encode_bytes(self.workstation_name)


class AvId(core.ValueEnum):
    MsvAvEOL = 0x0
    MsvAvNbComputerName = 0x1
    MsvAvNbDomainName = 0x2
    MsvAvDnsComputerName = 0x3
    MsvAvDnsDomainName = 0x4
    MsvAvDnsTreeName = 0x5
    MsvAvFlags = 0x6
    MsvAvTimestamp = 0x7
    MsvAvSingleHost = 0x8
    MsvAvTargetName = 0x9
    MsvChannelBindings = 0xA


AvId.import_items(globals())


class AvPair(core.Frame):
    text_fields = [
        MsvAvNbComputerName,
        MsvAvNbDomainName,
        MsvAvDnsComputerName,
        MsvAvDnsDomainName,
        MsvAvDnsTreeName,
        MsvAvTargetName,
    ]

    def __init__(self, parent=None):
        core.Frame.__init__(self, parent)
        if parent is not None:
            parent.target_info.append(self)
        self.av_id = None
        self.value = None

    def _encode(self, cur):
        cur.encode_uint16le(self.av_id)
        av_len_hole = cur.hole.encode_uint16le(0)
        av_start = cur.copy()
        if self.value is not None:
            if self.av_id in self.text_fields:
                cur.encode_utf16le(self.value)
            else:
                cur.encode_bytes(self.value)
        av_len_hole(cur - av_start)

    def _decode(self, cur):
        self.av_id = AvId(cur.decode_uint16le())
        av_len = cur.decode_uint16le()
        if av_len > 0:
            if self.av_id in self.text_fields:
                self.value = cur.decode_utf16le(av_len)
            else:
                self.value = cur.decode_bytes(av_len)


def extract_pair(av_pairs, av_id):
    for p in av_pairs:
        if p.av_id == av_id:
            return p


class NtLmChallengeMessage(core.Frame):
    message_type = NtLmChallenge

    def __init__(self, parent=None):
        core.Frame.__init__(self, parent)
        if parent is not None:
            parent.message = self
        self.target_name = None
        self.negotiate_flags = 0
        self.server_challenge = array.array("B", [0] * 8)
        self.target_info = []
        self.version = None

    def _decode(self, cur):
        message_start = cur.copy() - 12
        target_name_len = cur.decode_uint16le()
        target_name_max_len = cur.decode_uint16le()
        target_name_offset = cur.decode_uint32le()
        self.negotiate_flags = NegotiateFlags(cur.decode_uint32le())
        self.server_challenge = cur.decode_bytes(8)
        reserved = cur.decode_bytes(8)
        target_info_len = cur.decode_uint16le()
        target_info_max_len = cur.decode_uint16le()
        target_info_offset = cur.decode_uint32le()
        if self.negotiate_flags & NTLMSSP_NEGOTIATE_VERSION:
            with cur.bounded(cur, cur + 8):
                self.version = Version(self)
                self.version.decode(cur)
        else:
            cur.decode_uint64le()

        is_unicode = self.negotiate_flags & NTLMSSP_NEGOTIATE_UNICODE

        cur.seekto(message_start + target_name_offset)
        if target_name_len > 0:
            if is_unicode:
                self.target_name = cur.decode_utf16le(target_name_len)
            else:
                self.target_name = cur.decode_bytes(target_name_len)

        cur.seekto(message_start + target_info_offset)
        end = cur + target_info_len
        if target_info_len > 0:
            while cur < end:
                this_av = AvPair(self)
                with cur.bounded(cur, end):
                    this_av.decode(cur)
                if this_av.av_id == MsvAvEOL:
                    break


class NtLmAuthenticateMessage(core.Frame):
    message_type = NtLmAuthenticate

    def __init__(self, parent=None):
        core.Frame.__init__(self, parent)
        if parent is not None:
            parent.message = self
        self.lm_challenge_response = None
        self.nt_challenge_response = None
        self.domain_name = None
        self.user_name = None
        self.workstation_name = gethostname()
        self.encrypted_random_session_key = None
        self.negotiate_flags = None
        self.version = None
        self.mic = None

    def _encode(self, cur):
        is_unicode = self.negotiate_flags & NTLMSSP_NEGOTIATE_UNICODE

        message_start = cur - 8
        cur.encode_uint32le(self.message_type)

        if self.lm_challenge_response is not None:
            lm_challenge_response_len = len(self.lm_challenge_response)
        else:
            lm_challenge_response_len = 0
        cur.encode_uint16le(lm_challenge_response_len)
        cur.encode_uint16le(lm_challenge_response_len)
        lm_challenge_response_offset_hole = cur.hole.encode_uint32le(0)

        if self.nt_challenge_response is not None:
            nt_challenge_response_len = len(self.nt_challenge_response)
        else:
            nt_challenge_response_len = 0
        cur.encode_uint16le(nt_challenge_response_len)
        cur.encode_uint16le(nt_challenge_response_len)
        nt_challenge_response_offset_hole = cur.hole.encode_uint32le(0)

        if self.domain_name is not None:
            domain_name_len = len(self.domain_name)
        else:
            domain_name_len = 0
        if is_unicode:
            domain_name_len = domain_name_len * 2
        cur.encode_uint16le(domain_name_len)
        cur.encode_uint16le(domain_name_len)
        domain_name_offset_hole = cur.hole.encode_uint32le(0)

        if self.user_name is not None:
            user_name_len = len(self.user_name)
        else:
            user_name_len = 0
        if is_unicode:
            user_name_len = user_name_len * 2
        cur.encode_uint16le(user_name_len)
        cur.encode_uint16le(user_name_len)
        user_name_offset_hole = cur.hole.encode_uint32le(0)

        if self.workstation_name is not None:
            workstation_name_len = len(self.workstation_name)
        else:
            workstation_name_len = 0
        if is_unicode:
            workstation_name_len = workstation_name_len * 2
        cur.encode_uint16le(workstation_name_len)
        cur.encode_uint16le(workstation_name_len)
        workstation_name_offset_hole = cur.hole.encode_uint32le(0)

        if (
            self.negotiate_flags & NTLMSSP_NEGOTIATE_KEY_EXCH
            and self.encrypted_random_session_key is not None
        ):
            encrypted_random_session_key_len = len(self.encrypted_random_session_key)
        else:
            encrypted_random_session_key_len = 0
        cur.encode_uint16le(encrypted_random_session_key_len)
        cur.encode_uint16le(encrypted_random_session_key_len)
        encrypted_random_session_key_offset_hole = cur.hole.encode_uint32le(0)

        cur.encode_uint32le(self.negotiate_flags)

        if (
            self.negotiate_flags & NTLMSSP_NEGOTIATE_VERSION
            and self.version is not None
        ):
            self.version.encode(cur)
        else:
            cur.encode_uint64le(0)

        if self.mic is not None:
            cur.encode_bytes(self.mic[:16])

        lm_challenge_response_offset_hole(cur - message_start)
        if lm_challenge_response_len > 0:
            cur.encode_bytes(self.lm_challenge_response)

        nt_challenge_response_offset_hole(cur - message_start)
        if nt_challenge_response_len > 0:
            cur.encode_bytes(self.nt_challenge_response)

        domain_name_offset_hole(cur - message_start)
        if domain_name_len > 0:
            if is_unicode:
                cur.encode_utf16le(self.domain_name)
            else:
                cur.encode_bytes(self.domain_name.encode("ascii"))

        user_name_offset_hole(cur - message_start)
        if user_name_len > 0:
            if is_unicode:
                cur.encode_utf16le(self.user_name)
            else:
                cur.encode_bytes(self.user_name.encode("ascii"))

        workstation_name_offset_hole(cur - message_start)
        if workstation_name_len > 0:
            if is_unicode:
                cur.encode_utf16le(self.workstation_name)
            else:
                cur.encode_bytes(self.workstation_name)

        encrypted_random_session_key_offset_hole(cur - message_start)
        if encrypted_random_session_key_len > 0:
            cur.encode_bytes(self.encrypted_random_session_key)


class NtlmVersion(core.ValueEnum):
    NTLMv1 = 0x1
    NTLMv2 = 0x2


NtlmVersion.import_items(globals())


########################
## NTLMv1 Implementation
########################


def LMOWFv1(password):
    lm_passwd = password.upper() + "\0" * 14
    magic = "KGS!@#$%"
    return DES(lm_passwd[:7], magic) + DES(lm_passwd[7:14], magic)


def NTOWFv1(password):
    nt_passwd = password.encode("utf-16-le")
    return MD4.new(nt_passwd).digest()


def ComputeResponsev1(
    NegFlg,
    ResponseKeyNT,
    ResponseKeyLM,
    ServerChallenge,
    ClientChallenge,
    Time=None,
    ServerName=None,
):
    if NegFlg & NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY:
        NtChallengeResponse = DESL(
            ResponseKeyNT, MD5.new(ServerChallenge + ClientChallenge).digest()[:8]
        )
        LmChallengeResponse = ClientChallenge + "\0" * 16
    else:
        NtChallengeResponse = DESL(ResponseKeyNT, ServerChallenge)
        LmChallengeResponse = DESL(ResponseKeyLM, ServerChallenge)
    SessionBaseKey = MD4.new(ResponseKeyNT).digest()
    return NtChallengeResponse, LmChallengeResponse, SessionBaseKey


def KXKEY(NegFlg, SessionBaseKey, LmChallengeResponse, ServerChallenge, ResponseKeyLM):
    if NegFlg & NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY:
        hm = HMAC.new(SessionBaseKey, ServerChallenge + LmChallengeResponse[:8], MD5)
        KeyExchangeKey = hm.digest()
    else:
        LMOWF = ResponseKeyLM
        if NegFlg & NTLMSSP_NEGOTIATE_LMKEY:
            data = LmChallengeResponse[:8]
            KeyExchangeKey = DES(LMOWF[:7], data) + DES(LMOWF[8] + "\xbd" * 6, data)
        else:
            if NegFlg & NTLMSSP_REQUEST_NON_NT_SESSION_KEY:
                KeyExchangeKey = LMOWF[:8] + "\0" * 8
            else:
                KeyExchangeKey = SessionBaseKey
    return KeyExchangeKey


########################
## NTLMv2 Implementation
########################


class NTLMv2Response(core.Frame):
    def __init__(self, parent=None):
        core.Frame.__init__(self, parent)
        self.response = None
        self.challenge = None

    def _encode(self, cur):
        cur.encode_bytes(self.response)
        self.challenge.encode(cur)


class NTLMv2ClientChallenge(core.Frame):
    def __init__(self, parent=None):
        core.Frame.__init__(self, parent)
        if parent is not None:
            parent.challenge = self
        self.time_stamp = array.array("B", b"\0" * 8)
        self.challenge_from_client = array.array("B", b"\0" * 8)
        self.av_pairs = []

    def _encode(self, cur):
        cur.encode_uint8le(1)  # RespType
        cur.encode_uint8le(1)  # HiRespType
        cur.encode_uint16le(0)  # Reserved
        cur.encode_uint32le(0)  # Reserved
        cur.encode_uint64le(self.time_stamp)
        cur.encode_bytes(self.challenge_from_client)
        cur.encode_uint32le(0)  # Reserved
        for p in self.av_pairs:
            p.encode(cur)


def NTOWFv2(password, user, userdom):
    nt_passwd = password.encode("utf-16-le")
    nt_hash_passwd = MD4.new(nt_passwd).digest()
    nt_userdom = (user.upper() + userdom).encode("utf-16-le")
    return HMAC.new(nt_hash_passwd, nt_userdom, MD5).digest()


def ComputeResponsev2(
    NegFlg,
    ResponseKeyNT,
    ResponseKeyLM,
    ServerChallenge,
    ClientChallenge,
    Time=None,
    ServerName=None,
    av_pairs=None,
):
    if Time is None:
        Time = nttime.NtTime(nttime.datetime.now())
    if ServerName is None:
        ServerName = "SERVER"
    ServerName = ServerName.encode("utf-16-le")

    TimeBuf = array.array("B")
    cur = core.Cursor(TimeBuf, 0)
    cur.encode_uint64le(Time)

    Responseversion = "\x01"
    HiResponseversion = "\x01"

    ntlmv2_client_challenge = NTLMv2ClientChallenge()
    ntlmv2_client_challenge.time_stamp = Time
    ntlmv2_client_challenge.challenge_from_client = ClientChallenge
    if av_pairs is not None:
        ntlmv2_client_challenge.av_pairs = av_pairs
    temp = encode_frame(ntlmv2_client_challenge).tobytes()
    NTProofStr = HMAC.new(ResponseKeyNT, ServerChallenge + temp, MD5).digest()
    NtChallengeResponse = NTProofStr + temp
    LmChallengeResponse = (
        HMAC.new(ResponseKeyLM, ServerChallenge + ClientChallenge, MD5).digest()
        + ClientChallenge
    )
    SessionBaseKey = HMAC.new(ResponseKeyNT, NTProofStr, MD5).digest()
    return NtChallengeResponse, LmChallengeResponse, SessionBaseKey


class NtlmAuthenticator(object):
    """
    State machine for conducting ntlm authentication
    """

    neg_flags = (
        NTLMSSP_NEGOTIATE_UNICODE
        | NTLM_NEGOTIATE_OEM
        | NTLMSSP_REQUEST_TARGET
        | NTLMSSP_NEGOTIATE_NTLM
        | NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED
        | NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY
        | NTLM_NEGOTIATE_SIGN
        | NTLM_NEGOTIATE_SEAL
        | NTLMSSP_NEGOTIATE_128
        | NTLMSSP_NEGOTIATE_56
        | NTLMSSP_NEGOTIATE_KEY_EXCH
    )

    auth_flags = (
        NTLMSSP_NEGOTIATE_UNICODE
        | NTLMSSP_REQUEST_TARGET
        | NTLMSSP_NEGOTIATE_NTLM
        | NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY
        | NTLM_NEGOTIATE_SIGN
        | NTLM_NEGOTIATE_SEAL
        | NTLMSSP_NEGOTIATE_TARGET_INFO
        | NTLMSSP_NEGOTIATE_128
        | NTLMSSP_NEGOTIATE_56
        | NTLMSSP_TARGET_TYPE_DOMAIN
        | NTLMSSP_NEGOTIATE_KEY_EXCH
    )

    def __init__(self, domain, username, password):
        self.messages = []
        self.domain = domain
        self.username = username
        self.password = password
        self.version = Version()
        self.version.product_build = 9999
        self.ntlm_version = NTLMv2

        self.client_challenge = nonce(8)
        self.session_base_key = nonce(16)

        self.negotiate_message = None
        self.negotiate_buffer = None
        self.challenge_message = None
        self.challenge_buffer = None
        self.authenticate_message = None
        self.authenticate_buffer = None

    def ntlmv1(self):
        self.lm_hash = LMOWFv1(self.password.encode("ascii"))
        self.nt_hash = NTOWFv1(self.password)
        (
            self.nt_challenge_response,
            self.lm_challenge_response,
            self.session_base_key,
        ) = ComputeResponsev1(
            self.auth_flags,
            self.nt_hash,
            self.lm_hash,
            self.server_challenge.tobytes(),
            self.client_challenge.tobytes(),
        )
        self.key_exchange_key = KXKEY(
            self.auth_flags,
            self.session_base_key,
            self.lm_challenge_response,
            self.server_challenge.tobytes(),
            self.lm_hash,
        )

    def ntlmv2(self):
        ctarget_info = self.challenge_message.message.target_info
        server_time = extract_pair(ctarget_info, MsvAvTimestamp)
        if server_time is not None:
            time = nttime.NtTime(struct.unpack("<Q", server_time.value)[0])
        else:
            time = nttime.NtTime(nttime.datetime.now())
        server_name = extract_pair(ctarget_info, MsvAvNbComputerName)
        if server_name is not None:
            server_name = server_name.value

        self.nt_hash = NTOWFv2(self.password, self.username, self.domain)
        (
            self.nt_challenge_response,
            self.lm_challenge_response,
            self.session_base_key,
        ) = ComputeResponsev2(
            self.auth_flags,
            self.nt_hash,
            self.nt_hash,
            self.server_challenge.tobytes(),
            self.client_challenge.tobytes(),
            time,
            server_name,
            ctarget_info,
        )
        self.key_exchange_key = self.session_base_key

        if extract_pair(ctarget_info, MsvAvTimestamp) is not None:
            self.lm_challenge_response = b"\0" * 24

    def session_key(self):
        if self.auth_flags & NTLMSSP_NEGOTIATE_KEY_EXCH:
            r = RC4.new(self.key_exchange_key)
            self.exported_session_key = nonce(16)
            return r.encrypt(self.exported_session_key.tobytes())
        else:
            self.exported_session_key = self.key_exchange_key

    def negotiate(self):
        ntlm = Ntlm()
        neg = NtLmNegotiateMessage(ntlm)
        neg.negotiate_flags = self.neg_flags
        neg.domain_name = self.domain
        neg.version = self.version
        self.negotiate_buffer = array.array("B")
        ntlm.encode(core.Cursor(self.negotiate_buffer, 0))
        self.negotiate_message = ntlm
        return self.negotiate_buffer

    def authenticate(self, challenge_buf):
        # parse the challenge message
        self.challenge_buffer = challenge_buf
        ntlm_challenge = Ntlm()
        ntlm_challenge.decode(core.Cursor(challenge_buf, 0))
        self.server_challenge = ntlm_challenge.message.server_challenge
        self.challenge_message = ntlm_challenge

        # build the auth message
        ntlm = Ntlm()
        auth = NtLmAuthenticateMessage(ntlm)
        auth.negotiate_flags = self.auth_flags
        auth.version = self.version
        auth.domain_name = self.domain
        auth.user_name = self.username

        # perform the NTLM
        if self.ntlm_version == NTLMv1:
            self.ntlmv1()
        elif self.ntlm_version == NTLMv2:
            self.ntlmv2()
        else:
            raise NotImplementedError(
                "Unknown NTLM version {0}".format(self.ntlm_version)
            )

        # fill in the challenge responses
        auth.nt_challenge_response = self.nt_challenge_response
        auth.lm_challenge_response = self.lm_challenge_response

        # generate the session key if requested
        session_key = self.session_key()
        if session_key is not None:
            auth.encrypted_random_session_key = session_key

        self.authenticate_buffer = array.array("B")
        ntlm.encode(core.Cursor(self.authenticate_buffer, 0))
        self.authenticate_message = ntlm
        return self.authenticate_buffer



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        ioctl.py
#
# Abstract:
#
#        Test IOCTLs
#
# Authors: Ki Anderson (kimberley.anderson@emc.com)
#          Paul Martin (paul.o.martin@emc.com)
#          Steve Leef (steven.leef@dell.com)
#

import pike.model as model
import pike.smb2 as smb2
import pike.test as test
import pike.ntstatus


class ValidateNegotiateInfo(test.PikeTest):
    def __init__(self, *args, **kwds):
        super(ValidateNegotiateInfo, self).__init__(*args, **kwds)
        self.set_client_dialect(smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_0_2)

    # VALIDATE_NEGOTIATE_INFO fsctl succeeds for SMB3
    @test.RequireDialect(smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_0_2)
    def test_validate_negotiate_smb3(self):
        self.set_client_dialect(smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_0_2)
        chan, tree = self.tree_connect()
        chan.validate_negotiate_info(tree)


class QueryNetworkInterfaceInfo(test.PikeTest):
    def __init__(self, *args, **kwds):
        super(QueryNetworkInterfaceInfo, self).__init__(*args, **kwds)
        self.set_client_dialect(smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_1_1)

    @test.RequireDialect(smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_1_1)
    def test_query_interface_info_smb3(self):
        self.set_client_dialect(smb2.DIALECT_SMB3_0, smb2.DIALECT_SMB3_1_1)
        chan, tree = self.tree_connect()
        try:
            chan.query_network_interface_info(tree)
        except model.ResponseError as err:
            self.assertEqual(err.response.status, pike.ntstatus.STATUS_SUCCESS)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        query.py
#
# Abstract:
#
#        SMB2_QUERY_INFO command tests with SMB2_0_INFO_FILE type.
#
# Authors: Rafal Szczesniak (rafal.szczesniak@isilon.com)
#

import pike.model
import pike.ntstatus
import pike.smb2
import pike.test


class QueryTest(pike.test.PikeTest):
    def __init__(self, *args, **kwargs):
        super(QueryTest, self).__init__(*args, **kwargs)
        self.chan = None
        self.tree = None

    def open_file(self):
        """Helper to open basic file"""
        self.chan, self.tree = self.tree_connect()
        return self.chan.create(
            self.tree, "test.txt", disposition=pike.smb2.FILE_SUPERSEDE
        ).result()

    def basic_test(
        self, level, info_type=pike.smb2.SMB2_0_INFO_FILE, additional_information=None
    ):
        """Helper to perform a basic query test"""
        handle = self.open_file()
        info = self.chan.query_file_info(
            handle,
            level,
            info_type=info_type,
            additional_information=additional_information,
        )
        self.chan.close(handle)

    def mismatch_test(self, level, size):
        """Helper to perform a test size mismatch"""
        handle = self.open_file()

        with self.assert_error(pike.ntstatus.STATUS_INFO_LENGTH_MISMATCH):
            self.chan.query_file_info(handle, level, output_buffer_length=size)
        self.chan.close(handle)

    # The following tests simply try each info level
    def test_query_file_basic_info(self):
        self.basic_test(pike.smb2.FILE_BASIC_INFORMATION)

    def test_query_file_sec_info_owner(self):
        sec_info = pike.smb2.OWNER_SECURITY_INFORMATION
        self.basic_test(
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=sec_info,
        )

    def test_query_file_sec_info_group(self):
        sec_info = pike.smb2.GROUP_SECURITY_INFORMATION
        self.basic_test(
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=sec_info,
        )

    def test_query_file_sec_info_dacl(self):
        sec_info = pike.smb2.DACL_SECURITY_INFORMATION
        self.basic_test(
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=sec_info,
        )

    def test_query_file_sec_info_all(self):
        sec_info = (
            pike.smb2.OWNER_SECURITY_INFORMATION
            + pike.smb2.GROUP_SECURITY_INFORMATION
            + pike.smb2.DACL_SECURITY_INFORMATION
        )

        self.basic_test(
            pike.smb2.FILE_SECURITY_INFORMATION,
            info_type=pike.smb2.SMB2_0_INFO_SECURITY,
            additional_information=sec_info,
        )

    def test_query_file_standard_info(self):
        self.basic_test(pike.smb2.FILE_STANDARD_INFORMATION)

    def test_query_file_alignment_info(self):
        self.basic_test(pike.smb2.FILE_ALIGNMENT_INFORMATION)

    def test_query_file_internal_info(self):
        self.basic_test(pike.smb2.FILE_INTERNAL_INFORMATION)

    def test_query_file_mode_info(self):
        self.basic_test(pike.smb2.FILE_MODE_INFORMATION)

    def test_query_file_position_info(self):
        self.basic_test(pike.smb2.FILE_POSITION_INFORMATION)

    def test_query_file_ea_info(self):
        self.basic_test(pike.smb2.FILE_EA_INFORMATION)

    def test_query_file_all_info(self):
        self.basic_test(pike.smb2.FILE_ALL_INFORMATION)

    def test_mismatch_0_file_basic_info(self):
        self.mismatch_test(pike.smb2.FILE_BASIC_INFORMATION, 0)

    def test_mismatch_0_file_standard_info(self):
        self.mismatch_test(pike.smb2.FILE_STANDARD_INFORMATION, 0)

    def test_mismatch_0_file_alignment_info(self):
        self.mismatch_test(pike.smb2.FILE_ALIGNMENT_INFORMATION, 0)

    def test_mismatch_0_file_internal_info(self):
        self.mismatch_test(pike.smb2.FILE_INTERNAL_INFORMATION, 0)

    def test_mismatch_0_file_mode_info(self):
        self.mismatch_test(pike.smb2.FILE_MODE_INFORMATION, 0)

    def test_mismatch_0_file_position_info(self):
        self.mismatch_test(pike.smb2.FILE_POSITION_INFORMATION, 0)

    def test_mismatch_0_file_ea_info(self):
        self.mismatch_test(pike.smb2.FILE_EA_INFORMATION, 0)

    def test_mismatch_0_file_all_info(self):
        self.mismatch_test(pike.smb2.FILE_ALL_INFORMATION, 0)

    # Querying a security descriptor with output buffer size 0 should
    # return STATUS_BUFFER_TOO_SMALL and tell us how many bytes are needed
    def test_query_sec_desc_0(self):
        handle = self.open_file()
        smb2_req = self.chan.request(obj=handle)
        query_req = pike.smb2.QueryInfoRequest(smb2_req)
        query_req.info_type = pike.smb2.SMB2_0_INFO_SECURITY
        query_req.additional_information = (
            pike.smb2.OWNER_SECURITY_INFORMATION | pike.smb2.DACL_SECURITY_INFORMATION
        )
        query_req.output_buffer_length = 0
        query_req.file_id = handle.file_id

        with self.assert_error(pike.ntstatus.STATUS_BUFFER_TOO_SMALL) as cm:
            self.chan.connection.transceive(smb2_req.parent)

        err = cm.response[0]

        self.assertEqual(err[0].data_length, 4)
        self.assertTrue(err[0].minimum_buffer_length > 0)



import pytest

import pike.auth


# pike uses the "NONE" domain when one isn't provided
NONE = "NONE"


@pytest.mark.parametrize(
    "cred_str,exp_cred_tuple",
    (
        ("foo%bar", (NONE, "foo", "bar")),
        ("foo%ba%r", (NONE, "foo", "ba%r")),
        ("BAZ\\foo%bar", ("BAZ", "foo", "bar")),
        ("BAZ\\foo%b%ar", ("BAZ", "foo", "b%ar")),
        # not convinced this is valid, but it won't crash the splitter
        ("BA\\Z\\foo%bar", ("BA", "Z\\foo", "bar")),
        ("BA\\Z\\foo%bar%", ("BA", "Z\\foo", "bar%")),
    ),
)
def test_split_credentials(cred_str, exp_cred_tuple):
    assert pike.auth.split_credentials(cred_str) == exp_cred_tuple
